from main import app

from asynctor.testing import anyio_backend_fixture, async_client_fixture, chdir_tmp_fixture

anyio_backend = anyio_backend_fixture()
client = async_client_fixture(app)
tmp_work_dir = chdir_tmp_fixture()



from __future__ import annotations

from .xlsx import Excel, df_to_datas, pd_read_excel, read_excel
from .xlsx import load_xlsx as load_xls

# Leave it here for compatibility
__all__ = ("Excel", "df_to_datas", "load_xls", "pd_read_excel", "read_excel")



from __future__ import annotations

import sys
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Annotated

import pytest
import typer
from fastapi import FastAPI

from asynctor.compat import chdir
from asynctor.contrib.fastapi import RunServer, runserver
from asynctor.utils import get_machine_ip


@pytest.fixture
def mock_uvicorn_run(mocker):
    mock_object = mocker.patch("uvicorn.run")
    yield mock_object


@pytest.fixture
def mock_no_args(mocker):
    mocker.patch.object(sys, "argv", sys.argv[:1])
    yield


@dataclass
class CliOpts:
    addrport: Annotated[str | None, "Optional port number, or ipaddr:port"] = None
    port: int | None = None
    host: str = "0.0.0.0"
    reload: bool = False
    prod: bool = False
    verbose: bool = False

    def as_dict(self):
        return asdict(self)


def test_no_args(mock_uvicorn_run, mock_no_args):
    app = FastAPI()
    runserver(app)
    mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", reload=False)


class TestAsynctorPort:
    @pytest.fixture
    def mock_port(self, monkeypatch):
        monkeypatch.setenv("ASYNCTOR_PORT", "9000")
        yield

    def test_no_args_env_asynctor_port(self, mock_uvicorn_run, mock_no_args, mock_port):
        app = FastAPI()
        runserver(app)
        mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", port=9000, reload=False)

    def test_no_args_but_port_passed(self, mock_uvicorn_run, mock_no_args, mock_port):
        app = FastAPI()
        runserver(app, port=8888)
        mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", port=8888, reload=False)

    def test_with_args(self, mock_uvicorn_run, mock_port):
        opts = CliOpts(reload=True)
        RunServer.run(FastAPI(), echo=typer.secho, **opts.as_dict())
        mock_uvicorn_run.assert_called_once_with(
            "__main__:app", host="0.0.0.0", port=9000, reload=True
        )


class TestOpenBrowser:
    @pytest.fixture
    def mock_subprocess(self, mocker):
        mock_object = mocker.patch("subprocess.run")
        yield mock_object

    def test_no_args(self, mock_uvicorn_run, mock_subprocess, mock_no_args):
        app = FastAPI()
        runserver(app, open_browser=True)
        if sys.platform.lower().startswith("win"):
            mock_subprocess.assert_called_once_with(["explorer", "http://127.0.0.1:8000/docs"])
        else:
            mock_subprocess.assert_called_once_with(["open", "http://127.0.0.1:8000/docs"])
        mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", reload=False)

    def test_no_args_with_env(self, mock_uvicorn_run, mock_subprocess, mock_no_args, monkeypatch):
        monkeypatch.setenv("ASYNCTOR_BROWSER", "1")
        app = FastAPI()
        runserver(app)
        if sys.platform.lower().startswith("win"):
            mock_subprocess.assert_called_once_with(["explorer", "http://127.0.0.1:8000/docs"])
        else:
            mock_subprocess.assert_called_once_with(["open", "http://127.0.0.1:8000/docs"])
        mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", reload=False)

    def test_no_args_with_env_0(self, mock_uvicorn_run, mock_subprocess, mock_no_args, monkeypatch):
        monkeypatch.setenv("ASYNCTOR_BROWSER", "0")
        app = FastAPI()
        runserver(app)
        mock_subprocess.assert_not_called()
        mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", reload=False)

    def test_no_args_but_port_defined(self, mock_uvicorn_run, mock_subprocess, mock_no_args):
        app = FastAPI()
        runserver(app, port=9000, open_browser=True)
        if sys.platform.lower().startswith("win"):
            mock_subprocess.assert_called_once_with(["explorer", "http://127.0.0.1:9000/docs"])
        else:
            mock_subprocess.assert_called_once_with(["open", "http://127.0.0.1:9000/docs"])
        mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", port=9000, reload=False)

    def test_no_args_but_host_defined(self, mock_uvicorn_run, mock_subprocess, mock_no_args):
        app = FastAPI()
        runserver(app, host="192.168.0.3", open_browser=True)
        if sys.platform.lower().startswith("win"):
            mock_subprocess.assert_called_once_with(["explorer", "http://192.168.0.3:8000/docs"])
        else:
            mock_subprocess.assert_called_once_with(["open", "http://192.168.0.3:8000/docs"])
        mock_uvicorn_run.assert_called_once_with(app, host="192.168.0.3", reload=False)

    def test_with_args(self, mock_uvicorn_run, mock_subprocess, monkeypatch):
        monkeypatch.setenv("ASYNCTOR_BROWSER", "True")
        opts = CliOpts(reload=True)
        RunServer.run(FastAPI(), echo=typer.secho, **opts.as_dict())
        if sys.platform.lower().startswith("win"):
            mock_subprocess.assert_called_once_with(["explorer", "http://127.0.0.1:8000/docs"])
        else:
            mock_subprocess.assert_called_once_with(["open", "http://127.0.0.1:8000/docs"])
        mock_uvicorn_run.assert_called_once_with("__main__:app", host="0.0.0.0", reload=True)


def test_reload(mock_uvicorn_run):
    opts = CliOpts(reload=True)
    RunServer.run(FastAPI(), echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with("__main__:app", host="0.0.0.0", reload=True)


def test_host(mock_uvicorn_run):
    host = "127.0.0.1"
    _test_host(mock_uvicorn_run, host)


def _test_host(mock_uvicorn_run, host):
    app = FastAPI()
    opts = CliOpts(host=host)
    RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host=host, reload=False)


def test_host_ip(mock_uvicorn_run):
    host = get_machine_ip()
    _test_host(mock_uvicorn_run, host)


def test_host_domain(mock_uvicorn_run):
    host = "example.com"
    _test_host(mock_uvicorn_run, host)


def test_prod(mock_uvicorn_run):
    opts = CliOpts(prod=True)
    app = FastAPI()
    RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host=opts.host, reload=False)


def test_prod_hint(mock_uvicorn_run):
    opts = CliOpts(prod=True)
    app = FastAPI()
    with chdir(Path(__file__).parent):
        RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host=opts.host, port=9001, reload=False)


def test_port(mock_uvicorn_run):
    opts = CliOpts(port=9001)
    app = FastAPI()
    RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host=opts.host, port=9001, reload=False)


def test_addrport(mock_uvicorn_run):
    opts = CliOpts(addrport="9001")
    app = FastAPI()
    RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host=opts.host, port=9001, reload=False)


def test_addrport_localhost(mock_uvicorn_run):
    opts = CliOpts(addrport=":9001")
    app = FastAPI()
    RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host="127.0.0.1", port=9001, reload=False)


def test_addrport_zero_host(mock_uvicorn_run):
    opts = CliOpts(addrport="0:9001")
    app = FastAPI()
    RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host="0.0.0.0", port=9001, reload=False)


def test_addrport_ip_host(mock_uvicorn_run):
    host = get_machine_ip()
    opts = CliOpts(addrport=f"{host}:9001")
    app = FastAPI()
    RunServer.run(app, echo=typer.secho, **opts.as_dict())
    mock_uvicorn_run.assert_called_once_with(app, host=host, port=9001, reload=False)


def test_noreload(mocker):
    mocker.patch.object(sys, "argv", [sys.argv[0], "--noreload"])
    mock_object = mocker.patch("typer.run")
    runserver(FastAPI())
    assert sys.argv[1] == "--no-reload"
    mock_object.assert_called_once()



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeAlias

import anyio
import pandas as pd

if TYPE_CHECKING:
    from fastapi import UploadFile


FilePathType: TypeAlias = str | Path | anyio.Path


async def read_excel(
    file: UploadFile | FilePathType | BytesIO | bytes, as_str: bool = False, **kw
) -> pd.DataFrame:
    """Read excel from local file or bytes

    :param as_str: whether to read as dtype=str
    :param kw: other kwargs that will pass to the `pd.read_excel` function
    """
    if isinstance(file, (str, Path, BytesIO, bytes)):
        await anyio.lowlevel.checkpoint()
        return pd_read_excel(file, as_str, **kw)
    if isinstance(file, anyio.Path):
        content = await file.read_bytes()
    else:  # UploadFile
        content = await file.read()
    return pd_read_excel(content, as_str, **kw)


def pd_read_excel(
    file: str | Path | BytesIO | bytes, as_str: bool = False, *, keep_default_na: bool = False, **kw
) -> pd.DataFrame:
    if as_str and "dtype" not in kw:
        kw.setdefault("dtype", str)
    kw.update(keep_default_na=keep_default_na)
    return pd.read_excel(BytesIO(file) if isinstance(file, bytes) else file, **kw)


def df_to_datas(df: pd.DataFrame) -> list[dict]:
    """Convert dataframe to list of dict"""
    cols = list(df.columns)
    return [dict(zip(cols, v, strict=False)) for v in df.values.tolist()]


async def load_xlsx(
    file: UploadFile | FilePathType | BytesIO | bytes, as_str=False, **kw
) -> list[dict]:
    """Read excel file or content to be list of dict"""
    return df_to_datas(await read_excel(file, as_str, **kw))


class Excel:
    def __init__(self, filename: str | Path) -> None:
        self._path = Path(filename)

    @staticmethod
    def to_df(data: list[dict[str, Any]]) -> pd.DataFrame:
        columns = list(data[0].keys())
        df_data: dict[str, list[Any]] = {c: [] for c in columns}
        for d in data:
            for k, v in d.items():
                df_data[k].append(v)
        return pd.DataFrame(df_data)

    @classmethod
    def write_buffer(cls, data: list[dict[str, Any]] | pd.DataFrame) -> BytesIO:
        if not isinstance(data, pd.DataFrame):
            data = cls.to_df(data)
        buffer = BytesIO()
        with pd.ExcelWriter(buffer, engine="openpyxl") as writer:
            data.to_excel(writer, index=False)
        buffer.seek(0)
        return buffer

    def write(self, data: list[dict[str, Any]] | pd.DataFrame) -> None:
        df = data if isinstance(data, pd.DataFrame) else self.to_df(data)
        return df.to_excel(self._path, index=False)

    async def awrite(self, data: list[dict[str, Any]] | pd.DataFrame) -> None:
        bio = self.write_buffer(data)
        await anyio.Path(self._path).write_bytes(bio.getvalue())

    def read(self, as_str: bool = False, **kw) -> pd.DataFrame:
        return pd_read_excel(self._path, as_str, **kw)

    async def aread(self, as_str: bool = False, **kw) -> pd.DataFrame:
        return await read_excel(anyio.Path(self._path), as_str, **kw)



from __future__ import annotations

import functools
import itertools
import sys
import warnings
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator, Iterable, Sequence
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Literal, ParamSpec, TypeAlias, TypeVar, cast, overload

import anyio
from anyio import from_thread
from anyio._backends._asyncio import get_running_loop
from anyio.lowlevel import checkpoint

from .exceptions import ParamsError

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import TypeVarTuple, Unpack
else:
    from exceptiongroup import ExceptionGroup  # pragma: no cover
    from typing_extensions import TypeVarTuple, Unpack  # pragma: no cover
try:
    # anyio>=4.12.0 no longer depends on sniffio, and has it's own current_async_library
    from anyio._core._eventloop import current_async_library
except ImportError:
    import sniffio

    def current_async_library() -> str | None:
        try:
            asynclib_name = sniffio.current_async_library()
        except sniffio.AsyncLibraryNotFoundError:
            return None
        return asynclib_name


if TYPE_CHECKING:
    from anyio.abc._tasks import TaskGroup


T = TypeVar("T")
T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")
PosArgsT = TypeVarTuple("PosArgsT")
AwaitT = Awaitable[T_Retval]
CallableT = Callable[T_ParamSpec, AwaitT]
AsyncFunc: TypeAlias = Callable[..., Awaitable[Any]]
CoroFunc: TypeAlias = Awaitable[Any] | AsyncFunc


@overload
def ensure_afunc(coro: AwaitT) -> Callable[[], AwaitT]: ...


@overload
def ensure_afunc(coro: CallableT) -> CallableT: ...


def ensure_afunc(coro: AwaitT | CallableT) -> Callable[[], AwaitT] | CallableT:
    """Wrap coroutine to be async function"""
    if callable(coro):
        return coro

    async def do_await() -> T_Retval:
        return await coro

    return do_await


def run(
    func: Awaitable[T_Retval] | Callable[[Unpack[PosArgsT]], Awaitable[T_Retval]],
    *args: Unpack[PosArgsT],
    backend: str = "asyncio",
    backend_options: dict[str, Any] | None = None,
) -> T_Retval:
    """Combine `asyncio.run` and `anyio.run`

    :param func: async function or coroutine.
    :param args: arguments that will pass to `func` if it's a function.
    :param backend: should be 'asyncio' or 'trio'.
    :param backend_options: will pass to `anyio.run`.

    Usage::

    ... code-block:: python3

        from functools import partial

        async def foo(a, b, *, c=3):
            return a, b, c

        result_asyncio_format = run(foo(1, 2, c=0))
        result_anyio_format = run(partial(foo, c=0), 1, 2)
        assert result_asyncio_format == result_anyio_format

    """
    return anyio.run(be_awaitable(func), *args, backend=backend, backend_options=backend_options)


class LengthFixedList(list[T]):
    def append(self, _: T) -> None:
        raise TypeError(f"{self.__class__.__name__} is fixed-size")


async def map_group(
    func: AsyncFunc, todos: Iterable[Any], *, results: list[Any] | None = None
) -> None:
    """
    The `map_group` function asynchronously executes a given function for each set of arguments in a
    provided iterable using `anyio.create_task_group()`.

    :param func: The `func` parameter is an asynchronous function that will be applied to
        each item in the `todos` iterable
    :param todos: The `todos` parameter in the `map_group` function represents an iterable
        containing the arguments that will be passed to the `func` function for each task
        that is started in the task group. Each element in `todos` is a set of arguments
        that will be unpacked and passed to the `func`.
    :param results: The `results` parameter in the `map_group` function is a list that
        stores the results of the asynchronous function calls.
        If the `results` parameter is provided with an initial list,
        the results of each function call will be appended to this list.
        If `results` is None or not provided, tasks will run without any return.
    """
    async with anyio.create_task_group() as tg:
        should_append = results is not None and not isinstance(results, LengthFixedList)
        for args in todos:
            if should_append:
                cast(list[Any], results).append(None)
            tg.start_soon(func, *args)


class BulkGather:
    def __init__(self, results) -> None:
        self._results = results

    async def runner(self, index: int, coro: Awaitable[T_Retval]) -> None:
        self._results[index] = await coro

    async def limited_runner(
        self, index: int, coro: Awaitable[T_Retval], limiter: anyio.CapacityLimiter
    ) -> None:
        async with limiter:
            self._results[index] = await coro

    async def run(self, coros, batch_size, wait_last, total) -> None:
        is_generator = total == 0
        if not batch_size:
            await self.unlimit_gather(coros, is_generator)
        elif not wait_last:
            await self.run_in_capacity_limiter(batch_size, coros, is_generator)
        else:
            await self.only_start_new_one_after_last_batch_finish(
                batch_size, coros, is_generator, total
            )

    async def unlimit_gather(self, coros, is_generator) -> None:
        func = functools.partial(map_group, self.runner, enumerate(coros))
        if is_generator:
            func = functools.partial(func, results=self._results)
        await func()

    async def run_in_capacity_limiter(self, batch_size, coros, is_generator) -> None:
        limiter = anyio.CapacityLimiter(batch_size)
        todo_args = ((*item, limiter) for item in enumerate(coros))
        func = functools.partial(map_group, self.limited_runner, todo_args)
        if is_generator:
            func = functools.partial(func, results=self._results)
        await func()

    async def only_start_new_one_after_last_batch_finish(
        self, batch_size, coros, is_generator, total
    ) -> None:
        if is_generator:
            todos: list[tuple[int, Awaitable]] = []
            for count, args in zip(itertools.count(), enumerate(coros)):
                todos.append(args)
                self._results.append(None)
                if count % batch_size == 0:
                    await map_group(self.runner, todos)
                    todos.clear()
            if todos:
                await map_group(self.runner, todos)
        else:
            for start in range(0, total, batch_size):
                coros_slice = itertools.islice(coros, start, start + batch_size)
                args_items = ((start + i, coro) for i, coro in enumerate(coros_slice))
                await map_group(self.runner, args_items)

    @property
    def results(self):
        return tuple(self._results)


@overload
async def bulk_gather(
    coros: Sequence[Awaitable[T_Retval]] | Generator[Awaitable[T_Retval]],
    batch_size: int = 0,
    wait_last: bool = False,
    raises: Literal[True] = True,
    *,
    limit: int | None = None,
) -> tuple[T_Retval, ...]: ...


@overload
async def bulk_gather(
    coros: Sequence[Awaitable[T_Retval]] | Generator[Awaitable[T_Retval]],
    batch_size: int = 0,
    wait_last: bool = False,
    raises: Literal[False] = False,
    *,
    limit: int | None = None,
) -> tuple[T_Retval | None, ...]: ...


async def bulk_gather(
    coros: Sequence[Awaitable[T_Retval]] | Generator[Awaitable[T_Retval]],
    batch_size: int = 0,
    wait_last: bool = False,
    raises: bool = True,
    *,
    limit: int | None = None,
) -> tuple[T_Retval | None, ...]:
    """
    Similar like `asyncio.gather`, but support the `batch_size` parameter.
    If batch_size is not zero, running tasks will CapacityLimiter({batch_size}).

    :param coros: Coroutines
    :param batch_size: running tasks limit number, set 0 to be unlimit.
    :param wait_last: if True, wait last bulk tasks to complete then start new task group,
        else use anyio.CapacityLimiter to limit task number.
    :param raises: if True, raise Exception when coroutine failed, else return None.
    :param limit: (deprecated) only leave it here to compare with old version.
    """
    results: list[T_Retval | None]
    try:
        total = len(coros)  # type:ignore[arg-type]
    except TypeError:  # if coros is generator
        total = 0
        results = []
    else:
        if total == 0:
            await checkpoint()
            return ()
        results = LengthFixedList([None] * total)
    if limit is not None:
        if batch_size:
            if batch_size != limit:
                raise ParamsError(f"Conflict value with {limit=} & {batch_size=}")
            warnings.warn(
                "`limit` is deprecated, use `batch_size` only.",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            batch_size = limit
    bg = BulkGather(results)
    try:
        await bg.run(coros, batch_size, wait_last, total)
    except ExceptionGroup as e:
        if raises:
            raise e.exceptions[0] from e
    return bg.results


async def gather(*coros: Awaitable[T_Retval], limit: int | None = None) -> tuple[T_Retval, ...]:
    """Similar like asyncio.gather, but support set `limit` to control concurrency number.

    :param coros: Coroutines to be executed in the running loop
    :param limit: How many coroutines will be run in the same time (none or zero is unlimit).
    """
    return await bulk_gather(coros, limit=limit)


def create_task(coro: Awaitable[Any], task_group: TaskGroup, *, name: object = None) -> None:
    """Start an async task, similar to asyncio.create_task, but need a task_group param.

    :param coro: Coroutine that will be run in the backgroup
    :param task_group: The backgroup task group

    Usage::
        >>> async def afunc(): ...
        >>> async with anyio.create_task_group() as tg:
        ...     create_task(afunc(), tg)
    """
    task_group.start_soon(ensure_afunc(coro), name=name)


@asynccontextmanager
async def start_tasks(coro: CoroFunc, *more: CoroFunc) -> AsyncGenerator[None]:
    """Make it easy to convert asyncio.create_task

    Usage:

    ... code-block:: python3

        async def startup():
            # cost a long time to do sth async
            await anyio.sleep(1000)

        @contextlib.asynccontextmanager
        async def lifespan(app):
            async with start_tasks(startup()):
                yield
    """
    async with anyio.create_task_group() as tg:
        with anyio.CancelScope(shield=True):
            tg.start_soon(ensure_afunc(coro))
            for c in more:
                tg.start_soon(ensure_afunc(c))
            try:
                yield
            finally:
                tg.cancel_scope.cancel()


async def wait_for(coro: Awaitable[T_Retval], timeout: int | float) -> T_Retval:
    """Similar like asyncio.wait_for"""
    with anyio.fail_after(timeout):
        return await coro


def be_awaitable(
    async_func: AwaitT | Callable[[Unpack[PosArgsT]], AwaitT],
) -> Callable[[Unpack[PosArgsT]], AwaitT]:
    @functools.wraps(async_func)  # type:ignore[arg-type]
    async def do_await(*gs: Unpack[PosArgsT]) -> T_Retval:
        if callable(async_func):
            coro = async_func(*gs)
            try:
                res = await coro
            except TypeError as e:
                msg = (
                    "object can't be awaited"
                    if sys.version_info >= (3, 14)
                    else "can't be used in 'await' expression"
                )
                if msg in str(e):
                    # In case of async_func is a sync function
                    res = cast(T_Retval, coro)
                    await checkpoint()
                else:
                    raise
            return res
        else:
            if gs:
                warnings.warn("`run_async` for coroutine, does not handle 'args'.", stacklevel=1)
            return await async_func

    return do_await


def run_async(
    async_func: Awaitable[T_Retval] | Callable[[Unpack[PosArgsT]], Awaitable[T_Retval]],
    *args: Unpack[PosArgsT],
) -> T_Retval:
    """Run async function in worker thread and get the result of it"""
    # `asyncio.run(async_func())`/`anyio.run(async_func)` can get the result of async function,
    # but both of them will close the running loop.

    with from_thread.start_blocking_portal() as portal:
        future = portal.start_task_soon(be_awaitable(async_func), *args)
        return_value = future.result()

    return return_value


def async_to_sync(
    func: Callable[T_ParamSpec, Awaitable[T_Retval]],
) -> Callable[T_ParamSpec, T_Retval]:
    """Run async function to be sync.

    If there is a running loop, try to run in it,
    otherwise run in worker thread with new loop

    Usage::
        >>> async def do_sth():
        ...     await anyio.sleep(0.1)
        ...     return 1
        >>> res = async_to_sync(do_sth)()
        >>> assert res == 1
    """

    @functools.wraps(func)
    def runner(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> T_Retval:
        if current_async_library() == "asyncio":
            coro = func(*args, **kwargs)
            try:
                loop = get_running_loop()
                return loop.run_until_complete(coro)
            except RuntimeError as e:
                if "This event loop is already running" in str(e):
                    return run_async(ensure_afunc(coro))
                else:
                    raise e
        return run_async(functools.partial(func, **kwargs), *args)

    return runner


def run_until_complete(
    async_func: Awaitable[T_Retval] | Callable[[], Awaitable[T_Retval]],
) -> T_Retval:
    """Run async function or coroutine in runing loop or worker thread"""
    return async_to_sync(ensure_afunc(async_func))()



from __future__ import annotations

import functools
import logging
import os
import platform
import re
import sys
from collections.abc import AsyncGenerator, Callable
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Annotated, Any, TypeAlias

import uvicorn
from fastapi import Depends, FastAPI, Request
from fastapi.routing import _merge_lifespan_context

from ..client import AsyncRedis
from ..utils import Shell, load_bool


def register_aioredis(
    app: FastAPI,
    check_connection: bool = True,
    **kwargs: Annotated[Any, "Kwargs that will pass to `redis.asyncio.Redis.__init__`"],
) -> None:
    """
    Register redis to fastapi application

    :param app: the fastapi application instance
    :param check_connection: whether check redis server pingable
    :param kwargs: such like: host, port, etc.
    """

    @asynccontextmanager
    async def redis_lifespan(app_instance: FastAPI) -> AsyncGenerator[None]:
        async with AsyncRedis(app_instance, **kwargs):
            yield

    original_lifespan = app.router.lifespan_context
    app.router.lifespan_context = _merge_lifespan_context(redis_lifespan, original_lifespan)


async def get_redis_client(request: Request) -> AsyncRedis:
    return AsyncRedis(request)


AioRedis = Annotated[AsyncRedis, Depends(get_redis_client)]
AioRedisDep = AioRedis
AioRedisDep.__doc__ = """Get register redis cli from application.

Example::

    from asynctor.contrib.fastapi import AioRedisDep, register_aioredis
    from fastapi import FastAPI

    app = FastAPI()
    register_aioredis(app)

    @app.get('/')
    async def get_redis_keys(redis: AioRedisDep) -> list[str]:
        keys = await redis.keys()
        return keys

"""


def get_client_ip(request: Request) -> str:
    if x_forwarded_for := request.headers.get("x_forwarded_for"):
        # X-Forwarded-For may includes many IP address, the first one is origin IP
        return x_forwarded_for.split(",")[0]
    elif x_real_ip := request.headers.get("x_real_ip"):
        return x_real_ip
    elif forwarded := request.headers.get("forwarded"):
        # Forwarded: for=192.0.2.60;proto=http;by=203.0.113.43
        parts = forwarded.split(";")
        for part in parts:
            if part.startswith("for="):
                return part[4:]
    if request.client is None:  # pragma: no cover
        # request.client is not None if server started by uvicorn
        # put it here to improve type hints
        return ""
    return request.client.host


async def get_client_host(request: Request) -> str:
    # If the function is non-async and you use it as a dependency, it will run in a thread.
    # https://github.com/kludex/fastapi-tips?tab=readme-ov-file#9-your-dependencies-may-be-running-on-threads
    return get_client_ip(request)


ClientIpDep = Annotated[str, Depends(get_client_host)]
ClientIpDep.__doc__ = """Get real ip of request client.

Usage::
    >>> @app.get('/')
    >>> def index(client_ip: ClientIpDep):
    ...     assert isinstance(client_ip, str)
"""


def config_access_log_to_show_time(log: str = "uvicorn.access") -> None:
    """Config access logging format for uvicorn

    Usage::
        >>> from asynctor.contrib.fastapi import config_access_log_to_show_time
        >>> app = FastAPI()
        >>> config_access_log_to_show_time()
    """
    fmt = "%(asctime)s - %(levelname)s - %(message)s"
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(fmt))
    logging.getLogger(log).addHandler(handler)


PreStartFunc: TypeAlias = Callable[
    [
        Annotated[str, "host"],
        Annotated[int | None, "port"],
        Annotated[bool, "reload"],
        Annotated[dict[str, str] | None, "docs_params"],
    ],
    Any,
]


class RunServer:
    @staticmethod
    def uvicorn_run(app: FastAPI, host: str, port: int | None, reload: bool, **kw) -> None:
        asgi = "__main__:app" if reload else app
        run = functools.partial(uvicorn.run, asgi, host=host, reload=reload, **kw)
        if port:
            run(port=port)
        else:
            run()

    @staticmethod
    def parse_host_port(
        addrport: str | int, verbose: bool, echo: Callable
    ) -> tuple[str | None, int | None]:
        host, port = None, None
        if isinstance(addrport, int) or addrport.isdigit():
            port = int(addrport)
        elif ":" in addrport:
            h, p = addrport.split(":", 1)
            if not h:
                host = "127.0.0.1"
            elif h != "0":
                host = h
            if p.isdigit():
                port = int(p)
        elif verbose:
            echo(f"Ignore argument {addrport = }")
        return host, port

    @staticmethod
    def load_prod_port(config_file: Path, verbose: bool, echo: Callable) -> int:
        import importlib.util

        spec = importlib.util.spec_from_file_location(config_file.stem, config_file)
        if spec is not None and spec.loader is not None:
            gunicorn_config = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(gunicorn_config)
            if p := getattr(gunicorn_config, "PORT", 0):
                if verbose:
                    echo(f"Load `PORT = {p}` from {config_file}")
                return int(p)
            elif (p := getattr(gunicorn_config, "bind", "").split(":")[-1]).isdigit():
                if verbose:
                    echo(f"Load `bind = xxx:{p}` from {config_file}")
                return int(p)
            elif verbose:
                echo(f"{config_file} does not have 'PORT' attribute")
        elif verbose:
            echo(f"Failed to load module from {config_file}")
        return 0

    @staticmethod
    def echo_docs_url(
        app: FastAPI,
        host: str,
        port: int | None,
        docs_params: dict | None = None,
        echo: Callable | None = None,
    ) -> str:
        if host == "0.0.0.0":
            if declared_host := os.getenv("ASYNCTOR_HOST"):
                host = declared_host
            else:
                from asynctor.utils import get_machine_ip

                host = get_machine_ip()
        url = f"http://{host}:{port or 8000}{app.docs_url}"
        if docs_params:
            url += "?" + "&".join(f"{k}={v}" for k, v in docs_params.items())
        tip = "You can view docs at:"
        if echo is None:
            print(f"{tip}\n{url}")
        else:
            echo(tip)
            try:
                echo(url, bold=True)
            except TypeError:
                echo(url)
        return url

    @staticmethod
    def load_port_from_env() -> int | None:
        if p := os.getenv("ASYNCTOR_PORT"):
            try:
                return int(p)
            except ValueError:
                ...
        return None

    @classmethod
    def run(
        cls,
        app: FastAPI,
        addrport: str | None,
        port: int | None,
        host: str,
        reload: bool,
        prod: bool,
        verbose: bool,
        echo: Callable,
        docs_params: dict[str, str] | None = None,
        pre_start: PreStartFunc | None = None,
        open_browser: bool | None = None,
        **kw,
    ) -> None:
        if addrport:
            h, p = cls.parse_host_port(addrport, verbose, echo)
            if h:
                host = h
            if p:
                port = p
        elif prod:
            deployment_dir = Path("deployment")
            for level in range(5):
                if deployment_dir.exists():
                    if (gc := deployment_dir / "gunicorn_config.py").exists():
                        if _port := cls.load_prod_port(gc, verbose, echo):
                            port = _port
                    elif verbose:
                        echo(f"{gc.name} not found in {deployment_dir}")
                    break
                parent = Path.cwd().parent if level == 0 else deployment_dir.parent.parent
                deployment_dir = parent / deployment_dir.name
            else:
                if verbose:
                    echo(f"Deployment dir: {deployment_dir.name!r} not found")
        cls.echo_and_run(app, host, port, reload, docs_params, pre_start, echo, **kw)

    @classmethod
    def echo_and_run(cls, app, host, port, reload, docs_params, pre_start, echo=None, **kw) -> None:
        if not port:
            port = cls.load_port_from_env()
        url = cls.echo_docs_url(app, host, port, docs_params, echo)
        if pre_start is not None:
            try:
                pre_start(host=host, port=port, reload=reload, docs_params=docs_params)
            except TypeError:
                pre_start()
        if kw.pop("open_browser", False) or load_bool("ASYNCTOR_BROWSER"):
            command = "explorer" if platform.system() == "Windows" else "open"
            if host == "0.0.0.0" and (m := re.search(r"://(.*?)[:/]", url)):
                url = url.replace(m.group(1), "127.0.0.1")
            Shell([command, url]).run(verbose=True)
        cls.uvicorn_run(app, host, port, reload, **kw)


def runserver(
    app: FastAPI,
    addrport: str | int | None = None,
    port: int | None = None,
    host: str = "0.0.0.0",
    reload: bool = False,
    verbose: bool = False,
    docs_params: dict[str, str] | None = None,
    pre_start: PreStartFunc | None = None,
    open_browser: bool | None = None,
    **kw,
) -> None:
    kw.update(docs_params=docs_params, pre_start=pre_start, open_browser=open_browser)
    if not (args := sys.argv[1:]):
        return RunServer.echo_and_run(app, host, port, reload, **kw)
    try:
        import typer
    except ImportError:
        raise ImportError(
            "You must install typer or typer-slim to support arguments"
            ", e.g.: pip install typer-slim"
        ) from None

    def cli(
        addrport: Annotated[str | None, "Optional port number, or ipaddr:port"] = typer.Argument(
            default=None
        ),
        port: int | None = None,
        host: str = "0.0.0.0",
        reload: bool = False,
        prod: bool = False,
        verbose: bool = False,
    ) -> None:
        RunServer.run(app, addrport, port, host, reload, prod, verbose, echo=typer.secho, **kw)

    if (django_style_noreload := "--noreload") in args:
        sys.argv[sys.argv.index(django_style_noreload)] = "--no-reload"

    typer.run(cli)



from __future__ import annotations

import functools
from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal, TypeAlias, overload

from .exceptions import UnsupportedError

BasicJsonType: TypeAlias = str | bool | int | float | None
JsonAbleType: TypeAlias = dict[str, Any] | list[Any] | BasicJsonType
StrictJsonType: TypeAlias = dict[str, JsonAbleType] | list[JsonAbleType] | BasicJsonType
DumpedJsonType: TypeAlias = str | bytes | bytearray

try:
    import orjson
except ImportError:
    import json

    def json_dumps(
        obj: Any, default: Callable[[Any], Any] | None = None, *, pretty: bool = False
    ) -> str:
        dumps = functools.partial(json.dumps, obj, default=default)
        return dumps(indent=2, ensure_ascii=False) if pretty else dumps(separators=(",", ":"))

    def json_dump_bytes(
        obj: Any, default: Callable[[Any], Any] | None = None, *, pretty: bool = False
    ) -> bytes:
        return json_dumps(obj, pretty=pretty, default=default).encode()

    def json_loads(obj: DumpedJsonType) -> JsonAbleType:
        return json.loads(obj)
else:

    def json_dump_bytes(
        obj: Any, default: Callable[[Any], Any] | None = None, *, pretty: bool = False
    ) -> bytes:
        option = orjson.OPT_INDENT_2 if pretty else None
        return orjson.dumps(obj, option=option, default=default)

    def json_dumps(
        obj: Any, default: Callable[[Any], Any] | None = None, *, pretty: bool = False
    ) -> str:
        return json_dump_bytes(obj, pretty=pretty, default=default).decode()

    def json_loads(obj: DumpedJsonType) -> JsonAbleType:
        return orjson.loads(obj)


json_dump_bytes.__doc__ = """Serialize ``obj`` to a JSON formatted bytes

:param obj: object to be dumped
:param default: a function that return a serializable version of obj or raise TypeError
:param pretty: whether indent to humanize
"""
json_dumps.__doc__ = """Serialize ``obj`` to a JSON formatted string

:param obj: object to be dumped
:param default: a function that return a serializable version of obj or raise TypeError
:param pretty: whether indent to humanize
"""
json_loads.__doc__ = """Deserialize JSON to Python objects."""


class FastJson:
    @overload
    @classmethod
    def dumps(cls, obj: Any, output: Literal["str"], *, pretty: bool = False) -> str: ...
    @overload
    @classmethod
    def dumps(
        cls, obj: Any, output: Literal["bytes"] | Path = "bytes", *, pretty: bool = False
    ) -> bytes: ...

    @classmethod
    def dumps(
        cls,
        obj: Any,
        output: Literal["str", "bytes"] | Path = "bytes",
        *,
        pretty: bool = False,
    ) -> str | bytes:
        """Serialize ``obj`` to a JSON formatted string or bytes

        :param obj: object to be dumped.
        :param output: if 'str' return utf-8 string, elif 'bytes' return utf-8 encoded bytes,
                    elif Path write dumped content to it and return bytes, else raise ValueError.
        :param pretty: whether indent to humanize.
        """
        match output:
            case "bytes" | Path():
                content = json_dump_bytes(obj, pretty=pretty)
                if isinstance(output, Path):
                    output.write_bytes(content)
                return content
            case "str" | "string":
                return json_dumps(obj, pretty=pretty)
            case x if x is str:  # support: FastJson.dumps(obj, output=str)
                return json_dumps(obj, pretty=pretty)
            case _:
                raise UnsupportedError(f"Unsupported output format: {output!r}")

    loads = staticmethod(json_loads)



from __future__ import annotations

from contextlib import asynccontextmanager

import uvicorn
from fastapi import FastAPI, Request
from pydantic import BaseModel

from asynctor import AsyncRedis
from asynctor.contrib.fastapi import (
    AioRedis,
    AioRedisDep,
    ClientIpDep,
    config_access_log_to_show_time,
    register_aioredis,
)
from asynctor.utils import get_machine_ip


@asynccontextmanager
async def lifespan(fastapi_app: FastAPI):
    async with AsyncRedis(fastapi_app):
        yield


app = FastAPI()
app_for_utils_test = FastAPI(lifespan=lifespan)
app_default_to_mount_lifespan = FastAPI(lifespan=lifespan)
register_aioredis(app)
config_access_log_to_show_time()


@app.get("/")
async def root(redis: AioRedis) -> list[str]:
    return await redis.keys()


@app_default_to_mount_lifespan.get("/state")
@app_for_utils_test.get("/state")
async def state(request: Request):
    return {"redis": str(getattr(request.app.state, "redis", None))}


@app.get("/redis")
async def get_value_from_redis_by_key(redis: AioRedis, key: str) -> dict[str, str | None]:
    value = await redis.get(key)
    return {key: value and value.decode()}


class Item(BaseModel):
    key: str
    value: str


@app.post("/redis")
async def set_redis_key_value(redis: AioRedisDep, item: Item) -> dict[str, str | None]:
    await redis.set(item.key, item.value)
    return {item.key: (v := await redis.get(item.key)) and v.decode()}


@app.get("/ip")
def get_your_ip(client_ip: ClientIpDep) -> dict[str, str]:
    return {"client ip": client_ip, "server ip": get_machine_ip()}


if __name__ == "__main__":
    uvicorn.run(app)



from __future__ import annotations

import asyncio
import functools
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Any

import anyio
import pytest
from anyio.lowlevel import checkpoint

from asynctor.aio import (
    LengthFixedList,
    async_to_sync,
    bulk_gather,
    create_task,
    current_async_library,
    gather,
    run,
    run_async,
    run_until_complete,
    start_tasks,
    wait_for,
)
from asynctor.exceptions import ParamsError
from asynctor.timing import Timer


def test_run_async():
    async def foo(sth: Any = ""):
        return sth

    assert run_async(foo) == ""
    assert run_async(foo()) == ""
    assert run_async(foo(1)) == 1
    assert run_async(foo(None)) is None
    assert run_async(foo(foo)) == foo
    assert run_async(functools.partial(foo, 2)) == 2

    async def convert(s: Any) -> int:
        return int(s)

    with pytest.raises(TypeError):
        run_async(convert, None)
    coro = convert("1")
    with pytest.warns(UserWarning, match="`run_async` for coroutine, does not handle 'args'."):
        res = run_async(coro, 2)
    assert res == 1


def test_run():
    async def foo(sth: Any = ""):
        return sth

    assert run(foo) == ""
    assert run(foo()) == ""
    assert run(foo(1)) == 1
    assert run(foo(None)) is None
    assert run(foo(foo)) == foo
    assert run(functools.partial(foo, 2)) == 2
    assert run(foo, 2) == 2
    assert run(foo, 2, backend="asyncio") == 2
    assert run(foo, 2, backend="asyncio", backend_options=None) == 2


@pytest.mark.anyio
async def test_wait_for():
    async def do_sth(seconds=0.2):
        await anyio.sleep(seconds)
        return seconds

    assert (await wait_for(do_sth(), 1)) == 0.2
    with pytest.raises(TimeoutError):
        await wait_for(do_sth(), 0.1)


class MockServer:
    limit = 50
    count = 0
    OK = 200
    ERROR = 400

    @classmethod
    async def response(cls) -> int:
        cls.count += 1
        await anyio.sleep(0.1)
        status_code = cls.OK
        if cls.count > cls.limit:
            status_code = cls.ERROR
        cls.count -= 1
        return status_code


class TestGather:
    @pytest.mark.anyio
    async def test_gather(self):
        assert (await gather()) == ()

        async def a():
            return 1

        async def b():
            return "2"

        async def c():
            pass

        assert (await gather(a(), b(), c())) == (1, "2", None)

    @staticmethod
    async def raise_error_later(seconds, err_type):
        await anyio.sleep(seconds)
        raise err_type(f"{seconds = }")

    @staticmethod
    def is_the_same_error(e1, e2):
        return type(e1) is type(e2) and str(e1) == str(e2)

    def create_coros_for_raise(self):
        coro1 = self.raise_error_later(0.2, ValueError)
        coro2 = self.raise_error_later(0.1, AttributeError)
        coro3 = self.raise_error_later(0.11, OSError)
        return coro1, coro2, coro3

    @pytest.mark.anyio
    async def test_gather_raise(self):
        err_asyncio = err_anyio = None
        try:
            await gather(*self.create_coros_for_raise())
        except Exception as e:
            err_anyio = e
        if (backend := current_async_library()) == "asyncio":
            try:
                await asyncio.gather(*self.create_coros_for_raise())
            except Exception as e:
                err_asyncio = e
            assert self.is_the_same_error(err_asyncio, err_anyio)
        else:
            print(f"{backend = }")
            assert err_anyio is not None

    @pytest.mark.anyio
    async def test_gather_without_raise(self):
        results = await bulk_gather(self.create_coros_for_raise(), raises=False)
        assert results == (None, None, None)

    @pytest.mark.anyio
    async def test_bulk(self):
        total = 200
        with Timer("Use sema:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await bulk_gather(tasks, MockServer.limit)
            assert sum(i == MockServer.OK for i in results) == total
        with Timer("Without sema:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await bulk_gather(tasks, MockServer.limit, wait_last=True)
            assert all(i == MockServer.OK for i in results)
        with Timer("All start:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await gather(*tasks)
            assert any(i == MockServer.ERROR for i in results)

    @pytest.mark.anyio
    async def test_gather_limit(self):
        total = 200
        with Timer("Use sema:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await gather(*tasks, limit=MockServer.limit)
            assert sum(i == MockServer.OK for i in results) == total
        with Timer("All start:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await gather(*tasks, limit=0)
            assert any(i == MockServer.ERROR for i in results)

    @pytest.mark.anyio
    async def test_bulk_batch_size(self):
        total = 200
        with Timer("Use sema:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await bulk_gather(tasks, batch_size=MockServer.limit)
            assert sum(i == MockServer.OK for i in results) == total
        with Timer("Without sema:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await bulk_gather(tasks, batch_size=MockServer.limit, wait_last=True)
            assert all(i == MockServer.OK for i in results)

    @pytest.mark.anyio
    async def test_bulk_batch_size_generator(self):
        total = 200
        with Timer("Use sema(generator):"):
            tasks = (MockServer.response() for _ in range(total))
            results = await bulk_gather(tasks, batch_size=MockServer.limit)
            assert sum(i == MockServer.OK for i in results) == total
        with Timer("Without sema(generator):"):
            tasks = (MockServer.response() for _ in range(total))
            results = await bulk_gather(tasks, batch_size=MockServer.limit, wait_last=True)
            assert all(i == MockServer.OK for i in results)

    @pytest.mark.anyio
    async def test_bulk_limit(self):
        total = 200
        with Timer("Use sema:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await bulk_gather(tasks, limit=MockServer.limit)
            assert sum(i == MockServer.OK for i in results) == total
        with Timer("Without sema:"):
            tasks = [MockServer.response() for _ in range(total)]
            results = await bulk_gather(tasks, limit=MockServer.limit, wait_last=True)
            assert all(i == MockServer.OK for i in results)

    @pytest.mark.anyio
    async def test_bulk_limit_generator(self):
        total = 200
        with Timer("Use sema(generator):"):
            tasks = (MockServer.response() for _ in range(total))
            results = await bulk_gather(tasks, limit=MockServer.limit)
            assert sum(i == MockServer.OK for i in results) == total
        with Timer("Without sema(generator):"):
            tasks = (MockServer.response() for _ in range(total))
            results = await bulk_gather(tasks, limit=MockServer.limit, wait_last=True)
            assert all(i == MockServer.OK for i in results)

    @pytest.mark.anyio
    async def test_bulk_conflict_or_warning(self):
        tasks = [MockServer.response() for _ in range(200)]
        with pytest.raises(ParamsError):
            await bulk_gather(tasks, batch_size=MockServer.limit + 1, limit=MockServer.limit)
        with pytest.deprecated_call():
            await bulk_gather(tasks, batch_size=MockServer.limit, limit=MockServer.limit)

    @pytest.mark.anyio
    async def test_bulk_gather_generator(self):
        async def a():
            return 1

        lazy_tasks = (a() for _ in range(3))
        assert (await bulk_gather(lazy_tasks)) == (1, 1, 1)


class TestStartTasks:
    root = anyio.Path(__file__).parent
    names = ("tmp.txt", "tmp2.txt", "tmp3.txt")

    async def remove_files(self):
        for name in self.names:
            if await (p := self.root / name).exists():
                await p.unlink()

    async def startup(self):
        await self.root.joinpath(self.names[0]).touch()
        await anyio.sleep(1)
        await self.root.joinpath(self.names[1]).touch()

    async def running(self):
        while True:
            now = datetime.now()
            await self.root.joinpath(self.names[2]).write_text(str(now))
            await anyio.sleep(0.5)

    @pytest.mark.anyio
    async def test_start_tasks(self):
        await self.remove_files()
        root, names = self.root, self.names

        @asynccontextmanager
        async def lifespan(app):
            now = datetime.now()
            async with start_tasks(self.startup, self.running):
                await anyio.sleep(0.1)
                assert await root.joinpath(names[0]).exists()
                assert not (await root.joinpath(names[1]).exists())
                assert await root.joinpath(names[2]).exists()
                await anyio.sleep(1)
                assert await root.joinpath(names[0]).exists()
                assert await root.joinpath(names[1]).exists()
                assert (await root.joinpath(names[2]).read_text()) > str(now)
                yield
            now = datetime.now()
            assert (await root.joinpath(names[2]).read_text()) <= str(now)
            await self.remove_files()

        async with lifespan(None):
            for name in names:
                assert await root.joinpath(name).exists()

        for name in names:
            assert not await root.joinpath(name).exists()


def test_run_until_complete():
    async def append(container: list, element: Any) -> list:
        container.append(element)
        await anyio.sleep(0.01)
        return container

    a: list[Any] = []
    coro = append(a, 1)
    assert run_until_complete(coro) == [1]
    assert 1 in a
    afunc = functools.partial(append, a, 2)
    assert run_until_complete(afunc) == [1, 2]
    assert a == [1, 2]

    async def do_sth():
        with pytest.raises(RuntimeError):
            anyio.run(append, a, 3)
        coro = append(a, 3)
        with pytest.raises(RuntimeError):
            asyncio.run(coro)
        assert 3 not in a
        run_until_complete(coro)
        run_until_complete(functools.partial(append, a, 4))
        await append(a, 5)

    anyio.run(do_sth)
    assert a == [1, 2, 3, 4, 5]

    async def never():
        raise RuntimeError("Never ever")

    with pytest.raises(RuntimeError, match="Never ever"):
        run_until_complete(never())


def test_run_until_complete_function_arguments():
    def sync_func(): ...

    run_until_complete(sync_func)

    def async_func_requires_position_argument(a: int): ...

    with pytest.raises(TypeError):
        run_until_complete(async_func_requires_position_argument)  # type:ignore[arg-type]
    ints: list[int] = []

    def async_func_argument_with_default_value(a: int = 1):
        ints.append(a)

    run_until_complete(async_func_argument_with_default_value)
    assert ints == [1]

    with pytest.raises(TypeError):
        LengthFixedList([0]).append(1)


@pytest.mark.anyio
async def test_create_task():
    async def do_sth(seconds=0.01):
        await anyio.sleep(seconds)
        return seconds

    with Timer("Test concurrency", verbose=False) as t:
        async with anyio.create_task_group() as tg:
            create_task(do_sth(0.08), tg)
            create_task(do_sth(0.09), tg)
            create_task(do_sth(0.07), tg)
    assert 0.1 <= t.cost < 0.2


def test_async_to_sync():
    async def hi(who: Any = False):
        await checkpoint()
        if who:
            return f"Hello, {who}."
        raise RuntimeError("foo")

    with pytest.raises(RuntimeError, match="foo"):
        async_to_sync(hi)()

    assert async_to_sync(hi)("World") == "Hello, World."



from __future__ import annotations

from functools import cached_property
from io import BytesIO
from pathlib import Path
from typing import Any

import anyio
import pandas as pd
import pytest

from asynctor.xls import Excel, df_to_datas, load_xls, read_excel

try:
    from fastapi import UploadFile
except ImportError:

    class UploadFile:  # type:ignore[no-redef]
        def __init__(self, file):
            self.file = file

        async def read(self) -> bytes:
            return self.file.read()


@pytest.mark.anyio
async def test_read():
    demo = Path(__file__).parent / "demo.xlsx"
    content = demo.read_bytes()
    df = await read_excel(demo)
    df2 = await read_excel(content)
    df3 = await read_excel(anyio.Path(demo))
    df4 = await read_excel(BytesIO(content))
    df5 = await read_excel(UploadFile(BytesIO(content)))
    assert df2.compare(df).empty and df3.compare(df).empty
    assert df4.compare(df).empty and df5.compare(df).empty
    data = df_to_datas(df)
    assert data == [
        {"Column1": "row1-\\t%c", "Column2\nMultiLines": 0, "Column 3": 1, 4: ""},
        {"Column1": "r2c1\n00", "Column2\nMultiLines": "r2 c2", "Column 3": 2, 4: ""},
    ]
    assert data == (await load_xls(demo))

    assert (await load_xls(demo, True)) == [
        {"Column1": "row1-\\t%c", "Column2\nMultiLines": "0", "Column 3": "1", 4: ""},
        {"Column1": "r2c1\n00", "Column2\nMultiLines": "r2 c2", "Column 3": "2", 4: ""},
    ]


def test_init_excel():
    filename = "a.xlsx"
    assert Excel(filename)._path == Excel(Path(filename))._path


@pytest.mark.usefixtures("tmp_work_dir")
class TestExcel:
    file = Path("a.xlsx")
    df_data: dict[str, list[Any]] = {
        "产品": ["苹果", "香蕉", "橙子"],
        "销量": [100, 150, 120],
        "单价": [5.5, 3.2, 4.8],
    }

    @cached_property
    def klass(self) -> Excel:
        return Excel(self.file)

    @cached_property
    def data(self) -> list[dict[str, Any]]:
        columns = list(self.df_data)
        length = len(self.df_data[columns[0]])
        lines: list[dict[str, Any]] = [dict.fromkeys(columns) for _ in range(length)]
        for key, value in self.df_data.items():
            for i, v in enumerate(value):
                lines[i][key] = v
        return lines

    def test_write(self):
        assert not self.klass._path.exists()
        self.klass.write(self.data)
        assert self.data == self.klass.read().to_dict(orient="records")
        assert self.klass._path.exists()
        self.klass._path.unlink()
        self.klass.write(pd.DataFrame(self.df_data))
        assert self.data == self.klass.read().to_dict(orient="records")

    @pytest.mark.anyio
    async def test_awrite(self):
        assert not self.klass._path.exists()
        await self.klass.awrite(self.data)
        assert self.data == (await self.klass.aread()).to_dict(orient="records")
        assert self.klass._path.exists()
        self.klass._path.unlink()
        await self.klass.awrite(pd.DataFrame(self.df_data))
        assert self.data == (await self.klass.aread()).to_dict(orient="records")



import shlex
import shutil
import subprocess
from pathlib import Path

from asynctor.testing import tmp_workdir_fixture


def test_fixtures(tmp_work_dir):
    asset_dir = Path(__file__).parent / "assets" / "testing_fixtures"
    for p in asset_dir.glob("*.py"):
        dst = "conftest.py" if p.stem == "conftest_" else "."
        shutil.copy(p, dst)
    cmd = "pytest _tests.py"
    r = subprocess.run(shlex.split(cmd), capture_output=True, text=True, encoding="utf-8")
    assert r.returncode == 0
    assert "error" not in r.stdout.lower()


new_work_dir = tmp_workdir_fixture()
CURRENT_DIR = Path.cwd()


def test_pwd(new_work_dir):
    assert Path.cwd() != CURRENT_DIR
    assert not list(Path().glob("*"))
    temp_file = Path("a.txt")
    assert temp_file.write_bytes(b"1") == 1
    assert temp_file.read_text() == "1"
    temp_file.unlink()
    assert not temp_file.exists()



"""
This module is for those that support both (python >=3.9,<3.11) and (python >= 3.11)
"""

from __future__ import annotations

import os
import sys
import warnings
from typing import Any

__all__ = ("chdir", "StrEnum", "load_toml", "NotRequired", "Self")

if sys.version_info >= (3, 11):
    from contextlib import chdir
    from enum import StrEnum
    from typing import NotRequired, Self

    from tomllib import loads as load_toml
else:
    import contextlib
    from enum import Enum

    from typing_extensions import NotRequired, Self

    def load_toml(content: str) -> dict[str, Any]:
        try:
            import tomli
        except ImportError:
            tip = (
                "tomli is required for Python<3.11, you can install it by:"
                "\n\n  pip install tomli"
                '\n\nOr:\n\n  pip install "asynctor[toml]"\n'
            )
            warnings.warn(tip, stacklevel=2)
            raise

        return tomli.loads(content)

    class chdir(contextlib.AbstractContextManager):  # Copied from source code of Python3.13
        """Non thread-safe context manager to change the current working directory."""

        def __init__(self, path):
            self.path = path
            self._old_cwd = []

        def __enter__(self):
            self._old_cwd.append(os.getcwd())
            os.chdir(self.path)

        def __exit__(self, *excinfo):
            os.chdir(self._old_cwd.pop())

    # Copied from source code of Python3.14
    class ReprEnum(Enum):
        """
        Only changes the repr(), leaving str() and format() to the mixed-in type.
        """

    class StrEnum(str, ReprEnum):
        """
        Enum where members are also (and must be) strings
        """

        def __new__(cls, *values):
            "values must already be of type `str`"
            if len(values) > 3:
                raise TypeError("too many arguments for str(): %r" % (values,))
            if len(values) == 1:
                # it must be a string
                if not isinstance(values[0], str):
                    raise TypeError("%r is not a string" % (values[0],))
            if len(values) >= 2:
                # check that encoding argument is a string
                if not isinstance(values[1], str):
                    raise TypeError("encoding must be a string, not %r" % (values[1],))
            if len(values) == 3:
                # check that errors argument is a string
                if not isinstance(values[2], str):
                    raise TypeError("errors must be a string, not %r" % (values[2]))
            value = str(*values)
            member = str.__new__(cls, value)
            member._value_ = value
            return member

        @staticmethod
        def _generate_next_value_(name, start, count, last_values):
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()

        __str__ = str.__str__



from datetime import datetime
from pathlib import Path

import pytest

from asynctor.exceptions import AsynctorError, UnsupportedError
from asynctor.jsons import FastJson, json_dump_bytes, json_dumps, json_loads


def default(obj):
    if isinstance(obj, datetime):
        return str(obj)


def test_dumps():
    assert json_dumps({1: 1}) == '{"1":1}'
    assert json_dumps({1: 1}, pretty=True) == '{\n  "1": 1\n}'
    dt = datetime(2025, 11, 13)
    with pytest.raises(TypeError):
        json_dumps({1: dt})
    assert json_dumps({1: dt}, default=default) == '{"1":"2025-11-13 00:00:00"}'


def test_dump_bytes():
    assert json_dump_bytes({1: 1}) == b'{"1":1}'
    assert json_dump_bytes({1: 1}, pretty=True) == b'{\n  "1": 1\n}'


def test_loads():
    assert json_loads(b'{"1":1}') == {"1": 1}
    assert json_loads('{"1":1}') == {"1": 1}


def test_fast_json():
    assert FastJson.loads(b'{"1":1}') == {"1": 1}
    assert FastJson.loads.__doc__ == json_loads.__doc__
    assert FastJson.dumps({1: 1}, "str") == '{"1":1}'
    assert FastJson.dumps({1: 1}, "string") == '{"1":1}'  # type:ignore
    assert FastJson.dumps({1: 1}, output="str") == '{"1":1}'
    assert FastJson.dumps({1: 1}, output=str) == '{"1":1}'  # type:ignore
    assert FastJson.dumps({1: 1}, "str", pretty=True) == '{\n  "1": 1\n}'
    assert FastJson.dumps({1: 1}, pretty=True, output="str") == '{\n  "1": 1\n}'
    assert FastJson.dumps({1: 1}) == b'{"1":1}'
    assert FastJson.dumps({1: 1}, "bytes") == b'{"1":1}'
    assert FastJson.dumps({1: 1}, output="bytes") == b'{"1":1}'
    assert FastJson.dumps({1: 1}, pretty=True) == b'{\n  "1": 1\n}'
    assert FastJson.dumps({1: 1}, "str") == '{"1":1}'
    dt = datetime(2025, 11, 13)
    with pytest.raises(TypeError):
        FastJson.dumps({1: dt})
    # For `default` case, use `json_dumps` or `json_dump_bytes` instead
    with pytest.raises(TypeError):
        FastJson.dumps({1: dt}, default=default)  # type:ignore


def test_fast_json_unsupported_output():
    with pytest.raises(ValueError):
        FastJson.dumps({1: 1}, output=bytes)  # type:ignore
    with pytest.raises(AsynctorError):
        FastJson.dumps({1: 1}, output=bytes)  # type:ignore
    with pytest.raises(UnsupportedError):
        FastJson.dumps({1: 1}, output=bytes)  # type:ignore


def test_fast_json_write(tmp_work_dir):
    p = Path("a.json")
    assert FastJson.dumps({1: 1}, p) == b'{"1":1}'
    assert p.read_bytes() == b'{"1":1}'
    assert FastJson.dumps({1: 1}, p, pretty=True) == b'{\n  "1": 1\n}'
    assert p.read_bytes() == b'{\n  "1": 1\n}'
    assert FastJson.dumps({1: 1}, pretty=True, output=p) == b'{\n  "1": 1\n}'



from pathlib import Path

import pytest
from httpx import AsyncClient


@pytest.mark.anyio
async def test_index(client: AsyncClient):
    response = await client.get("/")
    assert response.json() == {"a": 1}


def test_current_dir_changed(tmp_work_dir):
    pwd = Path.cwd()
    assert pwd == tmp_work_dir
    root_dir = Path(__file__).parent
    assert pwd != root_dir
    assert not root_dir.is_relative_to(pwd)



import subprocess

from asynctor import __package__, __version__


def test_version():
    r = subprocess.run(["uv", "pip", "list"], capture_output=True)
    out = r.stdout.decode().strip()
    me = [j for i in out.splitlines() if (j := i.split())[0] == __package__][0]
    assert me[1] == __version__



from __future__ import annotations

from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from .compat import chdir
from .utils import AsyncClientGenerator, AsyncTestClient

if TYPE_CHECKING:
    from _pytest.fixtures import FixtureFunctionDefinition
    from fastapi import FastAPI


def async_client_fixture(app: FastAPI, mount_lifespan: bool = True) -> FixtureFunctionDefinition:
    """Async test client for FastAPI as session fixture

    :param app: a fastapi instance.
    :param mount_lifespan: if True, auto mount lifespan for app.

    Usage::

    ... code-block:: python3

        import pytest
        from asynctor.testing import async_client_fixture
        from httpx import AsyncClient

        from main import app

        client = async_client_fixture(app)

        @pytest.mark.anyio
        async def test_api(client: AsyncClient):
            response = await client.get("/")
            assert response.status_code == 200

    """

    @pytest.fixture(scope="session")
    async def client() -> AsyncClientGenerator:
        async with AsyncTestClient(app) as c:
            yield c

    return client


def anyio_backend_fixture() -> FixtureFunctionDefinition:
    @pytest.fixture(scope="session")
    def anyio_backend() -> str:
        return "asyncio"

    return anyio_backend


def tmp_workdir_fixture() -> FixtureFunctionDefinition:
    @pytest.fixture
    def tmp_work_dir(tmp_path: Path) -> Generator[Path]:
        with chdir(tmp_path):
            yield tmp_path

    return tmp_work_dir


chdir_tmp_fixture = tmp_workdir_fixture  # For backward compatible
